@@ -42,7 +42,7 @@ terms of the MIT license. A copy of the license can be found in the file
 #define mi_decl_cold
 #elif (defined(__GNUC__) && (__GNUC__ >= 3)) || defined(__clang__) // includes clang and icc
 #if !MI_TRACK_ASAN
-#define mi_decl_forceinline     __attribute__((always_inline))
+#define mi_decl_forceinline     __attribute__((always_inline)) inline
 #else
 #define mi_decl_forceinline     inline
 #endif
@@ -308,8 +308,7 @@ mi_msecs_t    _mi_clock_end(mi_msecs_t start);
 mi_msecs_t    _mi_clock_start(void);
 
 // "alloc.c"
-void*         _mi_page_malloc(mi_theap_t* theap, mi_page_t* page, size_t size) mi_attr_noexcept;                  // called from `_mi_theap_malloc_aligned`
-void*         _mi_page_malloc_zeroed(mi_theap_t* theap, mi_page_t* page, size_t size) mi_attr_noexcept;           // called from `_mi_theap_malloc_aligned`
+void*         _mi_page_malloc_zero(mi_theap_t* theap, mi_page_t* page, size_t size, bool zero) mi_attr_noexcept;                  // called from `_mi_theap_malloc_aligned`
 void*         _mi_theap_malloc_zero(mi_theap_t* theap, size_t size, bool zero, size_t* usable) mi_attr_noexcept;
 void*         _mi_theap_malloc_zero_ex(mi_theap_t* theap, size_t size, bool zero, size_t huge_alignment, size_t* usable) mi_attr_noexcept;     // called from `_mi_theap_malloc_aligned`
 void*         _mi_theap_realloc_zero(mi_theap_t* theap, void* p, size_t newsize, bool zero, size_t* usable_pre, size_t* usable_post) mi_attr_noexcept;
@@ -412,7 +411,7 @@ typedef struct mi_option_desc_s {
   Inlined definitions
 ----------------------------------------------------------- */
 #define MI_UNUSED(x)     (void)(x)
-#if (MI_DEBUG>0)
+#ifndef NDEBUG
 #define MI_UNUSED_RELEASE(x)
 #else
 #define MI_UNUSED_RELEASE(x)  MI_UNUSED(x)
@@ -1130,21 +1129,38 @@ static inline uintptr_t _mi_random_shuffle(uintptr_t x) {
 
 
 // ---------------------------------------------------------------------------------
-// Provide our own `_mi_memcpy` for potential performance optimizations.
+// Provide our own `_mi_memcpy/set` for potential performance optimizations.
 //
 // For now, only on x64/x86 we optimize to `rep movsb/stosb`.
-// Generally, we check for "fast short rep movsb" (FSRM) or "fast enhanced rep movsb" (ERMS) support
+// Generally, we check for "fast short rep movsb/stosb" (FSRM/FSRS) or "fast enhanced rep movsb" (ERMS) support
 // (AMD Zen3+ (~2020) or Intel Ice Lake+ (~2017). See also issue #201 and pr #253.
+// Todo: we see improvements on win32 but less with glibc; we might want to only enable this on windows.
 // ---------------------------------------------------------------------------------
 
-#if !MI_TRACK_ENABLED && defined(_WIN32) && (MI_ARCH_X64 || MI_ARCH_X86)
+#if !MI_TRACK_ENABLED && (MI_ARCH_X64 || MI_ARCH_X86) && (defined(_WIN32) || defined(__GNUC__))
 
-extern mi_decl_hidden bool _mi_cpu_has_fsrm;
-extern mi_decl_hidden bool _mi_cpu_has_erms;
+extern mi_decl_hidden size_t _mi_cpu_movsb_max;  // in init.c
+extern mi_decl_hidden size_t _mi_cpu_stosb_max;
+
+static inline void mi_rep_movsb(void* dst, const void* src, size_t n) {
+  #if defined(__GNUC__)
+  __asm volatile("rep movsb" : "+D"(dst), "+c"(n), "+S"(src) : : "memory");
+  #else
+  __movsb((unsigned char*)dst, (const unsigned char*)src, n);
+  #endif
+}
+
+static inline void mi_rep_stosb(void* dst, uint8_t val, size_t n) {
+  #if defined(__GNUC__)
+  __asm volatile("rep stosb" : "+D"(dst), "+c"(n) : "a"(val) : "memory");
+  #else
+  __stosb((unsigned char*)dst, val, n);
+  #endif
+}
 
 static inline void _mi_memcpy(void* dst, const void* src, size_t n) {
-  if ((_mi_cpu_has_fsrm && n <= 128) || (_mi_cpu_has_erms && n > 128)) {
-    __movsb((unsigned char*)dst, (const unsigned char*)src, n);
+  if mi_likely(n <= _mi_cpu_movsb_max) {  // has fsrm && n <= 127  (todo: and maybe has erms?)
+    mi_rep_movsb(dst, src, n);
   }
   else {
     memcpy(dst, src, n);
@@ -1152,54 +1168,14 @@ static inline void _mi_memcpy(void* dst, const void* src, size_t n) {
 }
 
 static inline void _mi_memset(void* dst, int val, size_t n) {
-  if ((_mi_cpu_has_fsrm && n <= 128) || (_mi_cpu_has_erms && n > 128)) {
-    __stosb((unsigned char*)dst, (uint8_t)val, n);
+  if mi_likely(n <= _mi_cpu_stosb_max) {  // has fsrs && n <= 127
+    mi_rep_stosb(dst, (uint8_t)val, n);
   }
   else {
     memset(dst, val, n);
   }
 }
 
-static inline void _mi_memset_small(void* dst, int val, size_t n) {
-  mi_assert_internal(n<=2*MI_SMALL_SIZE_MAX);
-  __stosb((unsigned char*)dst, (uint8_t)val, n);
-}
-
-#elif !MI_TRACK_ENABLED && defined(__GNUC__) && (MI_ARCH_X64 || MI_ARCH_X86)
-
-extern mi_decl_hidden bool _mi_cpu_has_fsrm;
-extern mi_decl_hidden bool _mi_cpu_has_erms;
-
-static inline void _mi_movsb(void* dst, const void* src, size_t n) {
-  __asm volatile("rep movsb" : "+D"(dst), "+c"(n), "+S"(src) : : "memory");
-}
-
-static inline void _mi_stosb(void* dst, uint8_t val, size_t n) {
-  __asm volatile("rep stosb" : "+D"(dst), "+c"(n) : "a"(val) : "memory");
-}
-
-static inline void _mi_memcpy(void* dst, const void* src, size_t n) {
-  if ((_mi_cpu_has_fsrm && n <= 128) || (_mi_cpu_has_erms && n > 128)) {
-    _mi_movsb(dst,src,n);
-  }
-  else {
-    memcpy(dst,src,n);
-  }
-}
-
-static inline void _mi_memset(void* dst, int val, size_t n) {
-  if ((_mi_cpu_has_fsrm && n <= 128) || (_mi_cpu_has_erms && n > 128)) {
-    _mi_stosb(dst, (uint8_t)val, n);
-  }
-  else {
-    memset(dst,val,n);
-  }
-}
-
-static inline void _mi_memset_small(void* dst, int val, size_t n) {
-  _mi_stosb(dst,val,n);
-}
-
 #else
 
 static inline void _mi_memcpy(void* dst, const void* src, size_t n) {
@@ -1210,11 +1186,6 @@ static inline void _mi_memset(void* dst, int val, size_t n) {
   memset(dst, val, n);
 }
 
-static inline void _mi_memset_small(void* dst, int val, size_t n) {
-  mi_assert_internal(n<=2*MI_SMALL_SIZE_MAX);
-  memset(dst, val, n);
-}
-
 #endif
 
 // -------------------------------------------------------------------------------
@@ -1238,12 +1209,6 @@ static inline void _mi_memset_aligned(void* dst, int val, size_t n) {
   _mi_memset(adst, val, n);
 }
 
-static inline void _mi_memset_aligned_small(void* dst, int val, size_t n) {
-  mi_assert_internal(n<=2*MI_SMALL_SIZE_MAX);
-  void* adst = __builtin_assume_aligned(dst, MI_INTPTR_SIZE);
-  _mi_memset_small(adst, val, n);
-}
-
 #else
 
 // Default fallback on `_mi_memcpy`
@@ -1257,31 +1222,16 @@ static inline void _mi_memset_aligned(void* dst, int val, size_t n) {
   _mi_memset(dst, val, n);
 }
 
-static inline void _mi_memset_aligned_small(void* dst, int val, size_t n) {
-  mi_assert_internal(n<=2*MI_SMALL_SIZE_MAX);
-  mi_assert_internal((uintptr_t)dst % MI_INTPTR_SIZE == 0);
-  _mi_memset_small(dst, val, n);
-}
-
 #endif
 
 static inline void _mi_memzero(void* dst, size_t n) {
   _mi_memset(dst, 0, n);
 }
 
-static inline void _mi_memzero_small(void* dst, size_t n) {
-  mi_assert_internal(n<=2*MI_SMALL_SIZE_MAX);
-  _mi_memset_small(dst, 0, n);
-}
-
 static inline void _mi_memzero_aligned(void* dst, size_t n) {
   _mi_memset_aligned(dst, 0, n);
 }
 
-static inline void _mi_memzero_aligned_small(void* dst, size_t n) {
-  mi_assert_internal(n<=2*MI_SMALL_SIZE_MAX);
-  _mi_memset_aligned_small(dst, 0, n);
-}
 
 
 #endif  // MI_INTERNAL_H
