@@ -77,6 +77,7 @@ static inline mi_bfield_t mi_bfield_all_set(void) {
 static inline mi_bfield_t mi_bfield_mask(size_t bit_count, size_t shiftl) {
   mi_assert_internal(bit_count > 0);
   mi_assert_internal(bit_count + shiftl <= MI_BFIELD_BITS);
+  mi_assert_internal(shiftl < MI_BFIELD_BITS);
   const mi_bfield_t mask0 = (bit_count < MI_BFIELD_BITS ? (mi_bfield_one() << bit_count)-1 : mi_bfield_all_set());
   return (mask0 << shiftl);
 }
@@ -1440,8 +1441,9 @@ bool _mi_bitmap_forall_setc_ranges(mi_bitmap_t* bitmap, mi_forall_set_fun_t* vis
 }
 
 // Visit all set bits in a bitmap but try to return ranges (within bfields) if possible,
-// but only in chunks of at least `rngslices` slices (that are also aligned at `rngslices`). 
+// but only in chunks of at least `rngslices` slices (that are also aligned at `rngslices`)
 // and clear those ranges atomically.
+// However, the `rngslices` are capped at `MI_BFIELD_BITS` at most.
 // Used by purging to purge larger ranges when possible. With transparent huge pages we only
 // want to purge whole huge pages (2 MiB) at a time which is what the `rngslices` parameter achieves.
 bool _mi_bitmap_forall_setc_rangesn(mi_bitmap_t* bitmap, size_t rngslices, mi_forall_set_fun_t* visit, mi_arena_t* arena, void* arg) 
@@ -1450,8 +1452,8 @@ bool _mi_bitmap_forall_setc_rangesn(mi_bitmap_t* bitmap, size_t rngslices, mi_fo
   if (rngslices<=1) {
     return _mi_bitmap_forall_setc_ranges(bitmap, visit, arena, arg);
   }
-  mi_assert_internal(rngslices <= MI_BFIELD_BITS);
-  if (rngslices > MI_BFIELD_BITS) { rngslices = MI_BFIELD_BITS;  }
+  // mi_assert_internal(rngslices <= MI_BFIELD_BITS);  
+  if (rngslices > MI_BFIELD_BITS) { rngslices = MI_BFIELD_BITS;  } // cap at MI_BFIELD_BITS at most
 
   // for all chunkmap entries
   const size_t chunkmap_max = _mi_divide_up(mi_bitmap_chunk_count(bitmap), MI_BFIELD_BITS);
@@ -1465,12 +1467,12 @@ bool _mi_bitmap_forall_setc_rangesn(mi_bitmap_t* bitmap, size_t rngslices, mi_fo
       mi_bchunk_t* const chunk = &bitmap->chunks[chunk_idx];
       for (size_t j = 0; j < MI_BCHUNK_FIELDS; j++) {
         const size_t base_idx = (chunk_idx*MI_BCHUNK_BITS) + (j*MI_BFIELD_BITS);
-        mi_bfield_t b = mi_atomic_exchange_relaxed(&chunk->bfields[j], 0);
-        mi_bfield_t skipped = 0;
-        mi_bfield_t rngmask = mi_bfield_mask(rngslices, 0);
-        do {
+        mi_bfield_t b = mi_atomic_exchange_relaxed(&chunk->bfields[j], 0);                // atomic clear
+        mi_bfield_t skipped = 0;                                                          // but track which bits we skip so we can restore them
+        for(size_t shift = 0; rngslices + shift <= MI_BFIELD_BITS; shift += rngslices) {  // per `rngslices` to keep alignment
+          const mi_bfield_t rngmask = mi_bfield_mask(rngslices, shift);
           if ((b & rngmask) == rngmask) {
-            const size_t idx = base_idx + mi_ctz(rngmask);
+            const size_t idx = base_idx + shift;
             if (!visit(idx, rngslices, arena, arg)) {
               // break early
               if (skipped != 0) {
@@ -1481,9 +1483,8 @@ bool _mi_bitmap_forall_setc_rangesn(mi_bitmap_t* bitmap, size_t rngslices, mi_fo
           }
           else {
             skipped = skipped | (b & rngmask);
-          }
-          rngmask <<= rngslices;
-        } while (rngmask != 0);
+          }          
+        } 
         
         if (skipped != 0) {
           mi_atomic_or_relaxed(&chunk->bfields[j], skipped);
@@ -1842,7 +1843,7 @@ bool mi_bbitmap_try_find_and_clearN_(mi_bbitmap_t* bbitmap, size_t tseq, size_t
 
   // iterate through the chunks
   size_t chunk_idx = 0;
-  while (chunk_idx < chunk_max - chunk_req)
+  while (chunk_idx <= chunk_max - chunk_req)
   {
     size_t count = 0;  // chunk count
     do {
