@@ -41,7 +41,8 @@ static void test_stl_allocators();
 static void test_join();              // issue #1177
 static void test_thread_leak(void);   // issue #1104
 static void test_perf(void);          // issue #1104
-
+static void test_perf2(void);         // issue #1104
+static void test_perf3(void);         // issue #1104
 
 #if _WIN32
 #include "main-override-dep.h"
@@ -59,7 +60,9 @@ int main() {
   // test_join();
 
   // test_thread_leak();
-  test_perf();
+  // test_perf();
+  test_perf2();
+  // test_perf3();
 
   //test_std_string();
   //test_thread_local();
@@ -540,3 +543,29 @@ void test_perf(void)
   test_perf_run();
   std::cout << "gsum: " << gsum.load() << "\n";
 }
+
+
+static int sum2;
+
+static void escape(uint8_t* p, size_t n) { 
+  p[std::rand() % n] = 42;
+  sum2 += p[std::rand() % n];
+}
+
+void test_perf2(void) {  
+  for (size_t i = 0; i < 100000000; i++) {
+    const size_t n = 1000;
+    uint8_t* p = (uint8_t*)calloc(1, n);
+    escape(p,n);
+    free(p);
+  }
+}
+
+void test_perf3(void) {
+  for (size_t i = 0; i < 5; i++) {
+    const size_t n = (size_t)1*1024*1024*1024;
+    uint8_t* p = (uint8_t*)calloc(1, n);
+    escape(p, n);
+    free(p);
+  }
+}
