@@ -1,5 +1,5 @@
 /* ----------------------------------------------------------------------------
-Copyright (c) 2018-2020 Microsoft Research, Daan Leijen
+Copyright (c) 2018-2025 Microsoft Research, Daan Leijen
 This is free software; you can redistribute it and/or modify it under the
 terms of the MIT license.
 -----------------------------------------------------------------------------*/
@@ -22,8 +22,12 @@ terms of the MIT license.
 #include <string.h>
 #include <assert.h>
 
-// #define MI_GUARDED
-// #define USE_STD_MALLOC
+// #define MI_GUARDED         1
+// #define USE_STD_MALLOC     1
+
+#ifndef USE_STD_MALLOC
+#define MI_USE_HEAPS       4
+#endif
 
 // > mimalloc-test-stress [THREADS] [SCALE] [ITER]
 //
@@ -40,7 +44,7 @@ static int ITER    = 20;
 static int THREADS = 8;
 static int SCALE   = 10;
 static int ITER    = 10;
-#elif 0
+#elif  0
 static int THREADS = 4;
 static int SCALE   = 10;
 static int ITER    = 20;
@@ -74,19 +78,31 @@ static size_t use_one_size = 0;               // use single object size of `N *
 static bool   main_participates = false;       // main thread participates as a worker too
 
 #ifdef USE_STD_MALLOC
+
 #define custom_calloc(n,s)    calloc(n,s)
 #define custom_realloc(p,s)   realloc(p,s)
 #define custom_free(p)        free(p)
+
 #else
+
 #include <mimalloc.h>
 #include <mimalloc-stats.h>
+
+#ifdef MI_USE_HEAPS
+static mi_heap_t* current_heap;
+#define custom_calloc(n,s)    mi_heap_calloc(current_heap,n,s)
+#define custom_realloc(p,s)   mi_heap_realloc(current_heap,p,s)
+#define custom_free(p)        mi_free(p)
+#else
 #define custom_calloc(n,s)    mi_calloc(n,s)
 #define custom_realloc(p,s)   mi_realloc(p,s)
 #define custom_free(p)        mi_free(p)
+#endif
 
 #ifndef NDEBUG
-#define xHEAP_WALK             // walk the heap objects?
+#define xMI_HEAP_WALK             // walk the theap objects?
 #endif
+
 #endif
 
 // transfer pointer between threads
@@ -137,7 +153,7 @@ static void* alloc_items(size_t items, random_t r) {
   }
   if (items>=32 && items<=40) items*=2;              // pthreads uses 320b allocations (this shows that more clearly in the stats)
   if (use_one_size > 0) items = (use_one_size / sizeof(uintptr_t));
-  if (items==0) items = 1;
+  if (items==0) items = 1;  
   uintptr_t* p = (uintptr_t*)custom_calloc(items,sizeof(uintptr_t));
   if (p != NULL) {
     for (uintptr_t i = 0; i < items; i++) {
@@ -162,9 +178,9 @@ static void free_items(void* p) {
   custom_free(p);
 }
 
-#ifdef HEAP_WALK
-static bool visit_blocks(const mi_heap_t* heap, const mi_heap_area_t* area, void* block, size_t block_size, void* arg) {
-  (void)(heap); (void)(area);
+#ifdef MI_HEAP_WALK
+static bool visit_blocks(const mi_theap_t* theap, const mi_theap_area_t* area, void* block, size_t block_size, void* arg) {
+  (void)(theap); (void)(area);
   size_t* total = (size_t*)arg;
   if (block != NULL) {
     *total += block_size;
@@ -217,10 +233,10 @@ static void stress(intptr_t tid) {
     }
   }
 
-  #ifdef HEAP_WALK
-  // walk the heap
+  #ifdef MI_HEAP_WALK
+  // walk the theap
   size_t total = 0;
-  mi_heap_visit_blocks(mi_heap_get_default(), true, visit_blocks, &total);
+  mi_theap_visit_blocks(mi_theap_get_default(), true, visit_blocks, &total);
   #endif
 
   // free everything that is left
@@ -238,27 +254,43 @@ static void stress(intptr_t tid) {
 static void run_os_threads(size_t nthreads, void (*entry)(intptr_t tid));
 
 static void test_stress(void) {
+  #ifdef MI_USE_HEAPS
+  mi_heap_t* prev_heaps[MI_USE_HEAPS] = { NULL };
+  #endif
   uintptr_t r = rand();
   for (int n = 0; n < ITER; n++) {
+    
+    #ifdef MI_USE_HEAPS
+    // new heap for each iteration
+    if (prev_heaps[MI_USE_HEAPS-1] != NULL) {
+      mi_heap_delete(prev_heaps[MI_USE_HEAPS-1]);   // delete from N iterations ago
+    }
+    for(int i = MI_USE_HEAPS-1; i > 0; i--) {
+      prev_heaps[i] = prev_heaps[i-1];
+    }
+    prev_heaps[0] = current_heap;
+    current_heap = mi_heap_new();
+    #endif  
+
     run_os_threads(THREADS, &stress);
+
     #if !defined(NDEBUG) && !defined(USE_STD_MALLOC)
     // switch between arena and OS allocation for testing
     // mi_option_set_enabled(mi_option_disallow_arena_alloc, (n%2)==1);
     #endif
-    #ifdef HEAP_WALK
+    #if defined(MI_HEAP_WALK) && defined(MI_USE_HEAPS)
     size_t total = 0;
-    mi_abandoned_visit_blocks(mi_subproc_main(), -1, true, visit_blocks, &total);
+    // mi_abandoned_visit_blocks(mi_subproc_main(), -1, true, visit_blocks, &total);
+    mi_heap_visit_blocks(heap, true, visit_blocks, &total);
     #endif
+
     for (int i = 0; i < TRANSFERS; i++) {
       if (chance(50, &r) || n + 1 == ITER) { // free all on last run, otherwise free half of the transfers
         void* p = atomic_exchange_ptr(&transfer[i], NULL);
         free_items(p);
       }
     }
-    #ifndef NDEBUG
-    //mi_collect(false);
-    //mi_debug_show_arenas(true);
-    #endif
+    
     #if !defined(NDEBUG) || defined(MI_TSAN)
     if ((n + 1) % 10 == 0) {
       printf("- iterations left: %3d\n", ITER - (n + 1));
@@ -270,7 +302,19 @@ static void test_stress(void) {
     }
     #endif
   }
-  // clean up
+  
+  #ifndef USE_STD_MALLOC
+  mi_stats_print(NULL);
+  #endif
+  
+  // clean up  (a bit too early to test the final free_items still works correctly)
+  #ifdef MI_USE_HEAPS
+  for (int i = 0; i < MI_USE_HEAPS; i++) {
+    mi_heap_delete(prev_heaps[i]); prev_heaps[i] = NULL;
+  }
+  mi_heap_delete(current_heap); current_heap = NULL;
+  #endif
+
   for (int i = 0; i < TRANSFERS; i++) {
     void* p = atomic_exchange_ptr(&transfer[i], NULL);
     if (p != NULL) {
@@ -312,12 +356,12 @@ int main(int argc, char** argv) {
   #ifdef MI_LINK_VERSION
     mi_version();
   #endif
-  #ifdef HEAP_WALK
+  #ifdef MI_HEAP_WALK
     mi_option_enable(mi_option_visit_abandoned);
   #endif
   #if !defined(NDEBUG) && !defined(USE_STD_MALLOC)
-    // mi_option_set(mi_option_arena_reserve, 32 * 1024 /* in kib = 32MiB */);
-    // mi_option_set(mi_option_purge_delay,1);
+    mi_option_set(mi_option_arena_reserve, 32 * 1024 /* in kib = 32MiB */);
+    mi_option_set(mi_option_purge_delay,1);
   #endif
   #if defined(NDEBUG) && !defined(USE_STD_MALLOC)
     // mi_option_set(mi_option_purge_delay,-1);
@@ -343,7 +387,11 @@ int main(int argc, char** argv) {
   if (SCALE > 100) {
     allow_large_objects = true;
   }
-  printf("Using %d threads with a %d%% load-per-thread and %d iterations %s\n", THREADS, SCALE, ITER, (allow_large_objects ? "(allow large objects)" : ""));
+  printf("Using %d threads with a %d%% load-per-thread and %d iterations%s", THREADS, SCALE, ITER, (allow_large_objects ? " (allow large objects)" : ""));
+  #if MI_USE_HEAPS
+  printf(" (using %d rolling heaps)", MI_USE_HEAPS);
+  #endif
+  printf("\n");
 
   #if !defined(NDEBUG) && !defined(USE_STD_MALLOC)
   mi_stats_reset();
@@ -366,13 +414,14 @@ int main(int argc, char** argv) {
 
 #ifndef USE_STD_MALLOC
   #ifndef NDEBUG
-  mi_debug_show_arenas();
   mi_collect(true);
-  char* json = mi_stats_get_json(0, NULL);
-  if (json != NULL) {
-    fputs(json,stderr);
-    mi_free(json);
-  }
+  mi_debug_show_arenas();
+  //mi_collect(true);
+  //char* json = mi_stats_get_json(0, NULL);
+  //if (json != NULL) {
+  //  fputs(json,stderr);
+  //  mi_free(json);
+  //}
   #endif
   mi_collect(true);
   mi_stats_print(NULL);
